"""
            uGecko
Python library for use with TCPGecko.
"""

__title__ = "ugecko"
__author__ = "WiiuDev, VCoding"
__license__ = "MIT"
__copyright__ = "Copyright 2015-present wiiudev, 2021-present VCoding"
__version__ = "1.3.0"

from .uGecko import *